#include <Evas.h>
#include <Ecore.h>
#include <Ecore_Evas.h>
#include <Emotion.h>

char *file = "dvd:/";
Ecore_Evas *ecore_evas;
Evas_Object *background;
Evas_Object *video;
Evas_Object *status;
Evas_Object *control;
Evas_Object *control_items[10] = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,};
int menu_visible =  0;
int menu_button = -1;
int control_menu_visible = 0;
int control_select = 0;

/* if the window manager requests a delete - quit cleanly */
static void
canvas_delete_request(Ecore_Evas *ecore_evas_in)
{
   ecore_main_loop_quit();
}

/* resize and move control menu objects based on current selection and size */
void control_layout(void)
{
   Evas_Coord max_w, max_h, x, y, w, h;
   int i, num;
   
   if (!control_menu_visible) return;

   /* find otu how many items in the menu and the maximum size of any item */
   num = 0;
   max_w = 0;
   max_h = 0;
   for (i = 0; i < 10; i++)
     {
	if (control_items[i])
	  {
	     evas_object_geometry_get(control_items[i], NULL, NULL, &w, &h);
	     if (w > max_w) max_w = w;
	     if (h > max_h) max_h = h;
	     num++;
	  }
     }
   /* get the current canvas size */
   evas_output_viewport_get(ecore_evas_get(ecore_evas), NULL, NULL, &w, &h);
   /* center the control menu */
   x = (w - max_w) / 2;
   y = (h - (max_h * num)) / 2;
   /* now lay out hte items */
   for (i = 0; i < 10; i++)
     {
	if (control_items[i])
	  {
	     evas_object_move(control_items[i], x, y);
	     /* if this item is the selected one, move the control object */
	     if (i == control_select)
	       {
		  evas_object_color_set(control_items[i], 255, 230, 50, 255);
		  evas_object_move(control, x, y);
		  evas_object_resize(control, max_w, max_h);
	       }
	     y += max_h;
	  }
     }
}

/* if the canvas is resized - resize the video too */
static void
canvas_resize(Ecore_Evas *ecore_evas_in)
{
   Evas_Coord x, y, w, h;
   int vw, vh;
   double ratio;
   
   /* get the current canvas size */
   evas_output_viewport_get(ecore_evas_get(ecore_evas), NULL, NULL, &w, &h);

   /* resize the black background rectangle */
   evas_object_move(background, 0, 0);
   evas_object_resize(background, w, h);

   /* resize the video object AND retain aspect ratio */
   /* get the video size in pixels */
   emotion_object_size_get(video, &vw, &vh);
   /* get the video ratio */
   ratio = emotion_object_ratio_get(video);
   /* if the ratio > 0.0 then the video wants awidth / height ratio on the */
   /* screen as returned indicated by this ratio value. this is irrespective */
   /* of the pixel size of the video and indicates the video wants to scale */
   if (ratio > 0.0)
     {
	x = 0;
	y = (h - (w / ratio)) / 2;
	if (y < 0)
	  {
	     y = 0;
	     x = (w - (h * ratio)) / 2;
	     w = h * ratio;
	  }
	else
	  h = w / ratio;
	evas_object_move(video, x, y);
	evas_object_resize(video, w, h);
     }
   else
     {
	if (vh > 0)
	  /* generate ratio from the pixel size */
	  ratio = (double)vw / (double)vh;
	else
	  ratio = 1.0;
	x = 0;
	y = (h - (w / ratio)) / 2;
	if (y < 0)
	  {
	     y = 0;
	     x = (w - (h * ratio)) / 2;
	     w = h * ratio;
	  }
	else
	  h = w / ratio;
	evas_object_move(video, 0, 0);
	evas_object_resize(video, w, h);
     }
   /* take care of the status object */
   evas_object_move(status, 20, 20);
   
   control_layout();
}

/* this is called when a key is pressed on the control menu */
void control_menu_key_press(void *data, Evas *canvas_in, Evas_Object *obj, void *event_info)
{
   /* the event type given to this callback is the Evas_Event_Key_Down */
   Evas_Event_Key_Down *ev;
   
   ev = (Evas_Event_Key_Down *)event_info;
   /* if we hit "Escape" or "q" lets quit */
   if ((!strcmp(ev->keyname, "Escape")) || 
       (!strcmp(ev->keyname, "q")))
     ecore_main_loop_quit();
   /* if we hit up arrow */
   else if (!strcmp(ev->keyname, "Up"))
     {
	/* reduce selection number */
	control_select--;
	/* if we try and go off the menu - reset our decrement */
	if (control_select < 0)
	  control_select++;
     }
   /* if we hit down arrow */
   else if (!strcmp(ev->keyname, "Down"))
     {
	/* increase selection number */
	control_select++;
	/* if we try and go off the menu, reverse our incriment */
	if ((control_select >= 10) || (control_items[control_select] == NULL))
	  control_select--;
     }
   else if (!strcmp(ev->keyname, "Return"))
     {
	int i;
	
	/* if we hit "return" we select the current item */
	emotion_object_spu_channel_set(video, control_select - 1);
	/* delete our control menu objects */
	evas_object_del(control);
	control = NULL;
	for (i = 0; i < 10; i++)
	  {
	     if (control_items[i])
	       {
		  evas_object_del(control_items[i]);
		  control_items[i] = NULL;
	       }
	  }
	/* reset flag for control menu */
	control_menu_visible = 0;
	/* set focus back to the background */
	evas_object_focus_set(background, 1);
     }
   /* re-layout the control menu. evas optimises this anyway for us so we */
   /* can just be dumb and move, resize etc. everything */
   control_layout();
}

/* create 1 item for the control menu */
void control_menu_item(int num, const char *label)
{
   control_items[num] = evas_object_text_add(ecore_evas_get(ecore_evas));
   evas_object_text_font_set(control_items[num], "Vera", 16);
   evas_object_text_text_set(control_items[num], label);
   evas_object_color_set(control_items[num], 255, 255, 255, 255);
   evas_object_pass_events_set(control_items[num], 1);
   evas_object_show(control_items[num]);
}

/* this is called to display a "control menu" to do things like change */
/* subtitle language, audio track, etc. */
void control_menu(void)
{
   int i, num;
   
   /* set the flag */
   control_menu_visible = 1;
   
   /* create a blue rectangle we will put behind things to hilight them */
   control = evas_object_rectangle_add(ecore_evas_get(ecore_evas));
   evas_object_event_callback_add(control, EVAS_CALLBACK_KEY_DOWN, control_menu_key_press, NULL);
   evas_object_color_set(control, 40, 60, 220, 180);
   evas_object_show(control);
   
   /* get the number of cpu (subtitle) channels */
   num = emotion_object_spu_channel_count(video);
   
   /* create a first itel of "off" that will map to -1 as the subtitle */
   control_menu_item(0, "OFF");
   
   /* now make 1 item for each of the languages */
   for (i = 1; ((i < (num+ 1)) && (i < 10)); i++)
     control_menu_item(i, emotion_object_spu_channel_name_get(video, i - 1));
   
   /* get the current spu channel, -1 == off */
   control_select = emotion_object_spu_channel_get(video) + 1;
   
   /* focus the control object to send keystrokes there */
   evas_object_focus_set(control, 1);
   
   /* fix layout of the control menu */
   control_layout();
}

/* this function is set to be called whenever a key press event happens */
/* on the object that has this callback set */
void key_press(void *data, Evas *canvas_in, Evas_Object *obj, void *event_info)
{
   /* the event type given to this callback is the Evas_Event_Key_Down */
   Evas_Event_Key_Down *ev;
   
   ev = (Evas_Event_Key_Down *)event_info;
   /* if we hit "Escape" or "q" lets quit */
   if ((!strcmp(ev->keyname, "Escape")) || 
       (!strcmp(ev->keyname, "q")))
     ecore_main_loop_quit();
   /* if we his "comma" (also the "<"  key on qwerty keyboards) rewind by */
   /* 10 seconds */
   else if (!strcmp(ev->keyname, "comma"))
     {
	/* if we arnet in menu mode then allow skipping around */
	if (!menu_visible)
	  {
	     double pos;
	     
	     pos = emotion_object_position_get(video);
	     emotion_object_position_set(video, pos - 10.0);
	  }
     }
   /* if we his "period" (also the ">"  key on qwerty keyboards) forward by */
   /* 10 seconds */
   else if (!strcmp(ev->keyname, "period"))
     {
	/* if we arne't in menu mode then allow skipping around */
	if (!menu_visible)
	  {
	     double pos;
	     
	     pos = emotion_object_position_get(video);
	     emotion_object_position_set(video, pos + 10.0);
	  }
     }
   /* allow our app to toggle "fullscreen" mode if you hit "f" */
   else if (!strcmp(ev->keyname, "f"))
     {
	if (!ecore_evas_fullscreen_get(ecore_evas))
	  ecore_evas_fullscreen_set(ecore_evas, 1);
	else
	  ecore_evas_fullscreen_set(ecore_evas, 0);
     }
   /* keyboard arrow keys */
   else if (!strcmp(ev->keyname, "Up"))
     {
	if (!menu_visible)
	  control_menu();
	else
	  emotion_object_event_simple_send(video, EMOTION_EVENT_UP);
     }
   else if (!strcmp(ev->keyname, "Down"))
     {
	if (!menu_visible)
	  control_menu();
	else
	  emotion_object_event_simple_send(video, EMOTION_EVENT_DOWN);
     }
   else if (!strcmp(ev->keyname, "Left"))
     {
	/* if no menu is around left will jump back 30 seconds */
	if (!menu_visible)
	  {
	     double pos;
	     
	     pos = emotion_object_position_get(video);
	     emotion_object_position_set(video, pos - 30.0);
	  }
	/* otherwise it will send a left event to the menu */
	else
	  emotion_object_event_simple_send(video, EMOTION_EVENT_LEFT);
     }
   else if (!strcmp(ev->keyname, "Right"))
     {
	/* if no menu is around left will jump forward 30 seconds */
	if (!menu_visible)
	  {
	     double pos;
	     
	     pos = emotion_object_position_get(video);
	     emotion_object_position_set(video, pos + 30.0);
	  }
	/* otherwise it will send a right event to the menu */
	else
	  emotion_object_event_simple_send(video, EMOTION_EVENT_RIGHT);
     }
   /* the "Return" key for selecting something something */
   else if (!strcmp(ev->keyname, "Return"))
     {
	/* if no menu is visible return can act as play/pause */
	if (!menu_visible)
	  {
	     if (emotion_object_play_get(video))
	       emotion_object_play_set(video, 0);
	     else
	       emotion_object_play_set(video, 1);
	  }
	else
	  emotion_object_event_simple_send(video, EMOTION_EVENT_SELECT);
     }
   /* the "M" or the "Menu" key */
   else if ((!strcmp(ev->keyname, "m")) ||
	    (!strcmp(ev->keyname, "Menu")))
     {
	/* this always tries to go to the main menu (menu1) */
	emotion_object_event_simple_send(video, EMOTION_EVENT_MENU1);
     }
   /* "Page Up" on the keyboard for previous chapter etc. */
   else if (!strcmp(ev->keyname, "Prior"))
     {
	/* if no menu is up then go to the previous chapter */
	if (!menu_visible)
	  emotion_object_event_simple_send(video, EMOTION_EVENT_PREV);
     }
   /* "Page Down" on the keyboard for next chapter etc. */
   else if (!strcmp(ev->keyname, "Next"))
     {
	/* if no menu is up then go to the next chapter */
	if (!menu_visible)
	  emotion_object_event_simple_send(video, EMOTION_EVENT_NEXT);
     }
}

/* our utility function to update the status object string */
void update_status(void)
{
   char buffer[512];
   double len, pos;
   int pos_h, pos_m, len_h, len_m;
   double pos_s, len_s;
   
   /* a menu is up - dont display time */
   if (menu_visible)
     evas_object_text_text_set(status, "(Menu)");
   else
     {
	/* get the current position and length (in seconds) */
	pos = emotion_object_position_get(video);
	len = emotion_object_play_length_get(video);
	/* now convert this into hrs:mins:secs (don't be scared - */
	/* it's only math!) */
	pos_h = (int)pos / (60 * 60);
	pos_m = ((int)pos / (60)) - (pos_h * 60);
	pos_s = pos - (pos_h * 60 * 60) - (pos_m * 60);
	len_h = (int)len / (60 * 60);
	len_m = ((int)len / (60)) - (len_h * 60);
	len_s = len - (len_h * 60 * 60) - (len_m * 60);
	/* print this to a stirng buffer */
	snprintf(buffer, sizeof(buffer), "%02i:%02i:%02.2f / %02i:%02i:%02.2f",
		 pos_h, pos_m, pos_s, len_h, len_m, len_s);
	evas_object_text_text_set(status, buffer);
     }
}

/* this function is called every time a video frame has been decoded and is */
/* ready for display */
void frame_decode(void *data, Evas_Object *obj, void *event_info)
{
   update_status();
}

/* this function is called every time the video stream changes size or */
/* aspect ratio */
void frame_resize(void *data, Evas_Object *obj, void *event_info)
{
   /* all we need to do is call our resize handler - that will rescale and */
   /* resize based on ratio and/or pixel sizes of the video stream */
   canvas_resize(ecore_evas);
}

/* this function is called every time the video changes length */
void length_change(void *data, Evas_Object *obj, void *event_info)
{
   update_status();
}
    
/* this function is called when the decoding stops */
void decode_stopped(void *data, Evas_Object *obj, void *event_info)
{
   emotion_object_position_set(video, 0.0);
   emotion_object_play_set(video, 1);
}

/* this function is called when the number of "menu" buttons changes just */
/* like with dvd's. this will be called as menus come and go */
void button_num_change(void *data, Evas_Object *obj, void *event_info)
{
   /* get the number of spu buttons as our "is a menu up" flag */
   menu_visible = emotion_object_spu_button_count_get(video);
   update_status();
}

/* this function is called when a button is selected (emotion allows the */
/* mouse to select buttons so you know if the mouse has selected one if */
/* this gets called without inputting any events). -1 means no button is */
/* selected */
void button_sel_change(void *data, Evas_Object *obj, void *event_info)
{
   menu_button = emotion_object_spu_button_get(video);
   update_status();
}

/* the main function of the program */
int main(int argc, char **argv)
{
   int i;
   int opt_fullscreen = 0;
   
   /* properly parse command-line arguments */
   for (i = 1; i < argc; i++)
     {
	/* start in fullscreen mode */
	if (!strcmp(argv[i], "-f"))
	  opt_fullscreen = 1;
	else
	  file = argv[i];
     }
   /* create a canvas, display it, set a title, callbacks to call on resize */
   /* or if the window manager asks it to be deleted */
   ecore_evas_init();
   ecore_evas = ecore_evas_software_x11_new(NULL, 0, 0, 0, 800, 600);
   ecore_evas_callback_delete_request_set(ecore_evas, canvas_delete_request);
   ecore_evas_callback_resize_set(ecore_evas, canvas_resize);   
   ecore_evas_title_set(ecore_evas, "My DVD Player");
   ecore_evas_name_class_set(ecore_evas, "my_dvd_player", "My_DVD_Player");
   /* if our start option said to run fullscreen, we should */
   if (opt_fullscreen) ecore_evas_fullscreen_set(ecore_evas, 1);
   ecore_evas_show(ecore_evas);

   /* lets add the current directory as a place to find font files */
   evas_font_path_append(ecore_evas_get(ecore_evas), ".");
   /* and have a font cache (512Kb) big enough to avoid re-loading fonts */
   evas_font_cache_set(ecore_evas_get(ecore_evas), 512 * 1024);
   /* lets also set an image cache to avoid re-loading images (4Mb) */
   evas_image_cache_set(ecore_evas_get(ecore_evas), 4 * 1024 * 1024);
   
   /* create a black background object */
   background = evas_object_rectangle_add(ecore_evas_get(ecore_evas));
   evas_object_color_set(background, 0, 0, 0, 255);
   evas_object_show(background);
   /* focuses the background so key events go to this object and call the */
   /* key_press function every time a key goes down */
   evas_object_focus_set(background, 1);
   evas_object_event_callback_add(background, EVAS_CALLBACK_KEY_DOWN, key_press, NULL);
   
   /* create a video object */
   video = emotion_object_add(ecore_evas_get(ecore_evas));
   
   /* set up callbacks to get "events" from the video object so we can */
   /* interact with the video stream as data is played */
   evas_object_smart_callback_add(video, "frame_decode", frame_decode, NULL);
   evas_object_smart_callback_add(video, "frame_resize", frame_resize, NULL);
   evas_object_smart_callback_add(video, "length_change", length_change, NULL);
   evas_object_smart_callback_add(video, "decode_stop", decode_stopped, NULL);
   evas_object_smart_callback_add(video, "button_num_change", button_num_change, NULL);
   evas_object_smart_callback_add(video, "button_change", button_sel_change, NULL);
   
   /* set the file and start playing and show the video object */
   emotion_object_file_set(video, file);
   emotion_object_play_set(video, 1);
   evas_object_show(video);

   /* lets put a simple text string on the top-left for displaying position */
   status = evas_object_text_add(ecore_evas_get(ecore_evas));
   /* set the font to "Vera" (.ttf is appended) at 16 pixels in size */
   evas_object_text_font_set(status, "Vera", 16);
   /* set the text string */
   evas_object_text_text_set(status, "Starting...");
   /* set the color to white (255, 255, 255) partially transparent (180) */
   evas_object_color_set(status, 255, 255, 255, 180);
   /* dont accept mouse events on this object - for display only */
   evas_object_pass_events_set(status, 1);
   evas_object_show(status);
   
   /* force an initial resize */
   canvas_resize(ecore_evas);
   
   /* run the main loop of the program - playing, drawing, handling events */
   ecore_main_loop_begin();
   
   /* if we exit the main loop we will shut down */
   ecore_evas_shutdown();
}
